#include <Windows.h>
#include "IOCTLWrapper.h"

#define   ERROR_IOCTL_FAILED   0xAA55

//#define   METHOD_BUFFERED      0L
//#define   FILE_READ_ACCESS     1L
//#define   FILE_WRITE_ACCESS    2L
#define   FILE_DEVICE_KRNLDRVR (0x80FF)

#define   IOCTL_QUERY_INFO     (0x80FF6000)
#define   IOCTL_PORT_IN        UInt32((FILE_DEVICE_KRNLDRVR << 16L) | (0x820 << 2L) | (METHOD_BUFFERED) | (FILE_READ_ACCESS  << 14L))
#define   IOCTL_PORT_OUT       UInt32((FILE_DEVICE_KRNLDRVR << 16L) | (0x822 << 2L) | (METHOD_BUFFERED) | (FILE_WRITE_ACCESS << 14L))
#define   IOCTL_MEM_IN         UInt32((FILE_DEVICE_KRNLDRVR << 16L) | (0x824 << 2L) | (METHOD_BUFFERED) | (FILE_READ_ACCESS  << 14L))
#define   IOCTL_MEM_OUT        UInt32((FILE_DEVICE_KRNLDRVR << 16L) | (0x826 << 2L) | (METHOD_BUFFERED) | (FILE_WRITE_ACCESS << 14L))

const PUCHAR IFKey = "System\\CurrentControlSet\\Control\\DeviceClasses\\{E77894A2-697F-11D0-958E-AF1FDADB2F1B}";



typedef struct
{
    UInt32 DeviceID;
    UInt16 BaseAddress;
    UInt16 Reserved;
} TCardInfo;

typedef struct
{
    UInt16 Addr;   // Address to read or write, in I/O space
    UInt8 Flags;   // Bit 0 means Addr is relative to the hDevice's base. Note that the "non-relative" device's base is 0, so in practice this is always set.
    UInt8 Size;    // number of significant bytes
    UInt32 Data;   // the data read/towrite
} TPortOp;

typedef struct
{
    HANDLE hDevice;
} TDeviceData;

SInt64 NumCards = 0;
TDeviceData Devices[256];
HANDLE hCentralDevice = 0;





//Internal support functions.

void LoadCardList(void)
{
    HKEY hIFKey = 0, hInterKey = 0, hSubKey = 0;
    PUCHAR *SubKey;
    UInt32 Status;

    NumCards = 0;

    Status = RegOpenKeyExA(HKEY_LOCAL_MACHINE, IFKey, 0, KEY_ENUMERATE_SUB_KEYS, &hIFKey);
    if ( Status != ERROR_SUCCESS ) return;

    for ( int I = 0; ; ++I )
    {
        char Nom[300];
        UInt64 ValueType, ValueSize, IntValue;
        Status = RegEnumKeyA(hIFKey, I, Nom, sizeof(Nom));
        if ( Status == ERROR_NO_MORE_ITEMS ) break;
        if ( Status != ERROR_SUCCESS ) continue;

        Status = RegOpenKeyExA(hIFKey, Nom, 0, KEY_QUERY_VALUE, &hInterKey);
        if ( Status != ERROR_SUCCESS ) continue;

        Status = RegOpenKeyExA(hInterKey, "\\#\\Control", 0, KEY_QUERY_VALUE, &hSubKey);
        if ( Status != ERROR_SUCCESS )
        {
            RegCloseKey(hInterKey);
            continue;
        }
        ValueSize = sizeof(IntValue);
        Status = RegQueryValueExA(hSubKey, "Linked", 0, &ValueType, (PUCHAR)&IntValue, &ValueSize);
        RegCloseKey(hSubKey);
        if ( Status != ERROR_SUCCESS )
        {
            RegCloseKey(hInterKey);
            continue;
        }
        if ( (ValueType != REG_DWORD) || ! IntValue )
        {
            RegCloseKey(hInterKey);
            continue;
        }

        Status = RegOpenKeyExA(hInterKey, "\\#", 0, KEY_QUERY_VALUE, &hSubKey);
        if ( Status != ERROR_SUCCESS )
        {
            RegCloseKey(hInterKey);
            continue;
        }
        ValueSize = sizeof(Nom);
        Status = RegQueryValueExA(hSubKey, "SymbolicLink", 0, &ValueType, Nom, &ValueSize);
        RegCloseKey(hSubKey);
        RegCloseKey(hInterKey);
        if ( Status != ERROR_SUCCESS ) continue;
        if ( (ValueType != REG_SZ) ) continue;

        Devices[NumCards].hDevice = CreateFile(Nom, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, 0);
        if ( Devices[NumCards].hDevice == 0 ) Devices[NumCards].hDevice = INVALID_HANDLE_VALUE;
        if ( Devices[NumCards].hDevice != INVALID_HANDLE_VALUE ) ++NumCards;
    }


}

int IsCardOpen(SInt32 CardNum)
{
    if ( CardNum < 0 ) return 0;
    if ( CardNum >= NumCards ) return 0;
    if ( Devices[CardNum].hDevice == INVALID_HANDLE_VALUE ) return 0;

    return -1;
}

int IsCentralOpen(void)
{
    if ( hCentralDevice == INVALID_HANDLE_VALUE ) return 0;

    return -1;
}

UInt16 _InPortB(HANDLE hDevice, UInt16 Address)
{
    TPortOp DataBuffer;
    UInt64 cbRet;

    DataBuffer.Addr  = Address;
    DataBuffer.Flags = 0x01;
    DataBuffer.Size  = 1;
    if ( DeviceIoControl(hDevice, IOCTL_PORT_IN, &DataBuffer, sizeof(DataBuffer), &DataBuffer, sizeof(DataBuffer), &cbRet, NULL) )
       return DataBuffer.Data & 0xFF;
    else
       return ERROR_IOCTL_FAILED
    ;
}

UInt16 _InPort(HANDLE hDevice, UInt16 Address)
{
    TPortOp DataBuffer;
    UInt64 cbRet;

    DataBuffer.Addr  = Address;
    DataBuffer.Flags = 0x01;
    DataBuffer.Size  = 2;
    if ( DeviceIoControl(hDevice, IOCTL_PORT_IN, &DataBuffer, sizeof(DataBuffer), &DataBuffer, sizeof(DataBuffer), &cbRet, NULL) )
       return DataBuffer.Data & 0xFFFF;
    else
       return ERROR_IOCTL_FAILED
    ;
}

UInt16 _InPortL(HANDLE hDevice, UInt16 Address)
{
    TPortOp DataBuffer;
    UInt64 cbRet;

    DataBuffer.Addr  = Address;
    DataBuffer.Flags = 0x01;
    DataBuffer.Size  = 4;
    if ( DeviceIoControl(hDevice, IOCTL_PORT_IN, &DataBuffer, sizeof(DataBuffer), &DataBuffer, sizeof(DataBuffer), &cbRet, NULL) )
       return DataBuffer.Data;
    else
       return ERROR_IOCTL_FAILED
    ;
}


UInt16 _OutPortB(HANDLE hDevice, UInt16 Address, UInt8 Value)
{
    TPortOp DataBuffer;
    UInt64 cbRet;

    DataBuffer.Addr  = Address;
    DataBuffer.Flags = 0x01;
    DataBuffer.Size  = 1;
    DataBuffer.Data  = Value;
    if ( DeviceIoControl(hDevice, IOCTL_PORT_OUT, &DataBuffer, sizeof(DataBuffer), &DataBuffer, sizeof(DataBuffer), &cbRet, NULL) )
       return 0;
    else
       return ERROR_IOCTL_FAILED
    ;
}

UInt16 _OutPort(HANDLE hDevice, UInt16 Address, UInt16 Value)
{
    TPortOp DataBuffer;
    UInt64 cbRet;

    DataBuffer.Addr  = Address;
    DataBuffer.Flags = 0x01;
    DataBuffer.Size  = 2;
    DataBuffer.Data  = Value;
    if ( DeviceIoControl(hDevice, IOCTL_PORT_OUT, &DataBuffer, sizeof(DataBuffer), &DataBuffer, sizeof(DataBuffer), &cbRet, NULL) )
        return 0;
    else
        return ERROR_IOCTL_FAILED
    ;
}

UInt16 _OutPortL(HANDLE hDevice, UInt16 Address, UInt32 Value)
{
    TPortOp DataBuffer;
    UInt64 cbRet;

    DataBuffer.Addr  = Address;
    DataBuffer.Flags = 0x01;
    DataBuffer.Size  = 4;
    DataBuffer.Data  = Value;
    if ( DeviceIoControl(hDevice, IOCTL_PORT_OUT, &DataBuffer, sizeof(DataBuffer), &DataBuffer, sizeof(DataBuffer), &cbRet, NULL) )
        return 0;
    else
        return ERROR_IOCTL_FAILED
    ;
}





void OpenAIOWDM(void)
{
    LoadCardList();
    hCentralDevice = CreateFile("\\\\.\\AIOWDMCore", GENERIC_WRITE | GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, 0);
    if ( hCentralDevice == 0 ) hCentralDevice = INVALID_HANDLE_VALUE;
}

void CloseAIOWDM(void)
{
    for ( int I = 0; I < NumCards; ++I )
    {
        if ( Devices[I].hDevice != INVALID_HANDLE_VALUE )
            CloseHandle(Devices[I].hDevice)
        ;
    }
    NumCards = 0;
    CloseHandle(hCentralDevice);
}

UInt64 GetNumCards(void)
{
    return NumCards;
}

UInt32 QueryCardInfo(SInt32 CardNum, UInt32 *pDeviceID, UInt32 *pBase)
{
    UInt64 cbRet;
    TCardInfo Data;

    if ( ! IsCardOpen(CardNum) ) return 0;

    ZeroMemory(&Data, sizeof(Data));
    if ( DeviceIoControl(Devices[CardNum].hDevice, IOCTL_QUERY_INFO, NULL, 0, &Data, sizeof(Data), &cbRet, NULL) )
    {
        if ( pDeviceID ) *pDeviceID = Data.DeviceID;
        if ( pBase     ) *pBase = Data.BaseAddress;
        return -1;
    }
    return 0;
}

UInt16 RelInPortB (SInt32 CardNum, UInt16 Offset) { if ( ! IsCardOpen(CardNum) ) return ERROR_IOCTL_FAILED; return _InPortB (Devices[CardNum].hDevice, Offset); }
UInt16 RelInPort  (SInt32 CardNum, UInt16 Offset) { if ( ! IsCardOpen(CardNum) ) return ERROR_IOCTL_FAILED; return _InPort  (Devices[CardNum].hDevice, Offset); }
UInt32 RelInPortL (SInt32 CardNum, UInt16 Offset) { if ( ! IsCardOpen(CardNum) ) return ERROR_IOCTL_FAILED; return _InPortL (Devices[CardNum].hDevice, Offset); }

UInt16 RelOutPortB (SInt32 CardNum, UInt16 Offset, UInt8  Value) { if ( ! IsCardOpen(CardNum) ) return ERROR_IOCTL_FAILED; return _OutPortB (Devices[CardNum].hDevice, Offset, Value); }
UInt16 RelOutPort  (SInt32 CardNum, UInt16 Offset, UInt16 Value) { if ( ! IsCardOpen(CardNum) ) return ERROR_IOCTL_FAILED; return _OutPort  (Devices[CardNum].hDevice, Offset, Value); }
UInt16 RelOutPortL (SInt32 CardNum, UInt16 Offset, UInt32 Value) { if ( ! IsCardOpen(CardNum) ) return ERROR_IOCTL_FAILED; return _OutPortL (Devices[CardNum].hDevice, Offset, Value); }

UInt16 InPortB (UInt16 Addr) { if ( ! IsCentralOpen() ) return ERROR_IOCTL_FAILED; return _InPortB (hCentralDevice, Addr); }
UInt16 InPort  (UInt16 Addr) { if ( ! IsCentralOpen() ) return ERROR_IOCTL_FAILED; return _InPort  (hCentralDevice, Addr); }
UInt32 InPortL (UInt16 Addr) { if ( ! IsCentralOpen() ) return ERROR_IOCTL_FAILED; return _InPortL (hCentralDevice, Addr); }

UInt16 OutPortB(UInt16 Addr, UInt8  Value) { if ( ! IsCentralOpen() ) return ERROR_IOCTL_FAILED; return _OutPortB (hCentralDevice, Addr, Value); }
UInt16 OutPort (UInt16 Addr, UInt16 Value) { if ( ! IsCentralOpen() ) return ERROR_IOCTL_FAILED; return _OutPort  (hCentralDevice, Addr, Value); }
UInt16 OutPortL(UInt16 Addr, UInt32 Value) { if ( ! IsCentralOpen() ) return ERROR_IOCTL_FAILED; return _OutPortL (hCentralDevice, Addr, Value); }

